@@ -1,4 +1,4 @@
-// Copyright (c) 2010-2023, Lawrence Livermore National Security, LLC. Produced
+// Copyright (c) 2010-2024, Lawrence Livermore National Security, LLC. Produced
 // at the Lawrence Livermore National Laboratory. All Rights reserved. See files
 // LICENSE and NOTICE for details. LLNL-CODE-806117.
 //
@@ -176,11 +176,12 @@ TEST_CASE("SerialConstrainedSolver", "[ConstrainedSolver]")
 // marked [Parallel] because it uses hypre
 TEST_CASE("ConstrainedSolver", "[Parallel], [ConstrainedSolver]")
 {
-#ifdef HYPRE_USING_GPU
-   mfem::out << "\nAs of mfem-4.3 and hypre-2.22.0 (July 2021) this unit test\n"
-             << "is NOT supported with the GPU version of hypre.\n\n";
-   return;
-#endif
+   if (HypreUsingGPU())
+   {
+      mfem::out << "\nAs of mfem-4.3 and hypre-2.22.0 (July 2021) this unit test\n"
+                << "is NOT supported with the GPU version of hypre.\n\n";
+      return;
+   }
 
    int comm_size;
    MPI_Comm_size(MPI_COMM_WORLD, &comm_size);
@@ -381,11 +382,12 @@ void ParallelTestProblem::Penalty(double pen, Vector& serr, Vector& lerr)
 /// *actual* parallel constrained solver
 TEST_CASE("ParallelConstrainedSolver", "[Parallel], [ConstrainedSolver]")
 {
-#ifdef HYPRE_USING_GPU
-   mfem::out << "\nAs of mfem-4.3 and hypre-2.22.0 (July 2021) this unit test\n"
-             << "is NOT supported with the GPU version of hypre.\n\n";
-   return;
-#endif
+   if (HypreUsingGPU())
+   {
+      mfem::out << "\nAs of mfem-4.3 and hypre-2.22.0 (July 2021) this unit test\n"
+                << "is NOT supported with the GPU version of hypre.\n\n";
+      return;
+   }
 
    int comm_size;
    MPI_Comm_size(MPI_COMM_WORLD, &comm_size);
@@ -671,11 +673,12 @@ void ParallelTestProblemTwo::Penalty(double pen, Vector& serr, Vector& lerr)
 
 TEST_CASE("ParallelConstrainedSolverTwo", "[Parallel], [ConstrainedSolver]")
 {
-#ifdef HYPRE_USING_GPU
-   mfem::out << "\nAs of mfem-4.3 and hypre-2.22.0 (July 2021) this unit test\n"
-             << "is NOT supported with the GPU version of hypre.\n\n";
-   return;
-#endif
+   if (HypreUsingGPU())
+   {
+      mfem::out << "\nAs of mfem-4.3 and hypre-2.22.0 (July 2021) this unit test\n"
+                << "is NOT supported with the GPU version of hypre.\n\n";
+      return;
+   }
 
    int comm_rank, comm_size;
    MPI_Comm_rank(MPI_COMM_WORLD, &comm_rank);
@@ -811,11 +814,12 @@ void ZerosTestProblem::Elimination(Vector& serr, Vector& lerr, bool twoblocks)
 
 TEST_CASE("ZerosTestCase", "[Parallel], [ConstrainedSolver]")
 {
-#ifdef HYPRE_USING_GPU
-   mfem::out << "\nAs of mfem-4.3 and hypre-2.22.0 (July 2021) this unit test\n"
-             << "is NOT supported with the GPU version of hypre.\n\n";
-   return;
-#endif
+   if (HypreUsingGPU())
+   {
+      mfem::out << "\nAs of mfem-4.3 and hypre-2.22.0 (July 2021) this unit test\n"
+                << "is NOT supported with the GPU version of hypre.\n\n";
+      return;
+   }
 
    int comm_rank, comm_size;
    MPI_Comm_rank(MPI_COMM_WORLD, &comm_rank);
